#include <stdio.h>
#include "gps.h"
#include <string.h>
#include <pthread.h>
#include <memory.h>
#include <time.h>
#include <errno.h>
#include <unistd.h>

static const GpsInterface* sGpsInterface = NULL;
//static const GpsXtraInterface* sGpsXtraInterface = NULL;
//static const AGpsInterface* sAGpsInterface = NULL;
//static const GpsNiInterface* sGpsNiInterface = NULL;
static const GpsDebugInterface* sGpsDebugInterface = NULL;
//static const AGpsRilInterface* sAGpsRilInterface = NULL;

#define WAKE_LOCK_NAME  "GPS"

pthread_mutex_t location_lock;
pthread_mutex_t status_lock;
pthread_mutex_t sv_lock;
pthread_mutex_t nmea_lock;

int setTimeAndExit	= 0;




int timespec2str(char *buf, uint len, struct timespec *ts) {
    int ret;
    struct tm t;

    tzset();
    if (localtime_r(&(ts->tv_sec), &t) == NULL)
        return 1;

    ret = strftime(buf, len, "%F %T", &t);
    if (ret == 0)
        return 2;
    len -= ret - 1;

    ret = snprintf(&buf[strlen(buf)], len, ".%09ld", ts->tv_nsec);
    if (ret >= len)
        return 3;

    return 0;
}



static void checkAndClearExceptionFromCallback(const char* methodName) 
{
#if DEBUG	
	printf("%s\n", methodName);
#else
	
#endif	
}

/**
 * NmeaString structure.
 */ 
typedef struct {	
	/**
	 * Allocates space for individual NMEA string.
	 */ 
	char string[100]; 
	
	/**
	 * Actual NMEA string length
	 */ 
	int length;
	
	/**
	 * Timestamp of the NMEA string. 
	 * It does not work with GPS driver version 1.0,
	 * as it is always returned as 1.
	 * 
	 */ 
	GpsUtcTime timestamp;
} NmeaString;

/**
 * Storage for GPS callbacks.
 * The callback write to it, and the display thread 
 * reads from it.
 */ 
typedef struct {

	/**
	 * running display thread flag
	 */
	int runThread;

	/**
	 * display thread
	 */
	pthread_t               thread;
	
	/**
	 * gps location storage
	 */
	GpsLocation 						location;
	
	/**
	 * gps satellite vehicle (SV) storage
	 */ 
	GpsSvStatus							sv_status;
	
	/**
	 * gps status storage
	 */ 
	GpsStatus								status;
	
	/**
	 * Raw NMEA sentences storage 
	 */ 
	NmeaString							nmeaStrings[7];
	
	/**
	 * Tracks the storage index of the NMEA raw string.
	 */  
	int											nmeaStringIndex;
        
} GpsState;

/**
 * Instance of the GPS callbacks storage
 */ 
static GpsState  _gps_state[1];

/**
 * Callback with location information.
 */ 
static void location_callback(GpsLocation* location)
{	
	checkAndClearExceptionFromCallback(__FUNCTION__);
	
	pthread_mutex_lock(&location_lock);
	memcpy(&_gps_state->location, location, sizeof(_gps_state->location));
	pthread_mutex_unlock(&location_lock);
}

/**
 * Callback with status information.
 */ 
static void status_callback(GpsStatus* status)
{
	checkAndClearExceptionFromCallback(__FUNCTION__);
	
	pthread_mutex_lock(&status_lock);
	memcpy(&_gps_state->status, status, sizeof(_gps_state->status));
	pthread_mutex_unlock(&status_lock);		
}

/**
 * Callback with SV status information.
 */ 
static void sv_status_callback(GpsSvStatus* sv_status)
{
	checkAndClearExceptionFromCallback(__FUNCTION__);
		
	pthread_mutex_lock(&sv_lock);
	memcpy(&_gps_state->sv_status, sv_status, sizeof(_gps_state->sv_status));
	pthread_mutex_unlock(&sv_lock);
}

/**
 * Callback for reporting NMEA sentences.
 */ 
static void nmea_callback(GpsUtcTime timestamp, const char* nmea, int length)
{
	checkAndClearExceptionFromCallback(__FUNCTION__);
	
	pthread_mutex_lock(&nmea_lock);
	// reset the sentence we write into
	bzero(_gps_state->nmeaStrings[_gps_state->nmeaStringIndex].string, sizeof(_gps_state->nmeaStrings[_gps_state->nmeaStringIndex].string));
	
	// copy the NMEA sentence
	strncpy(_gps_state->nmeaStrings[_gps_state->nmeaStringIndex].string, nmea, length);
	_gps_state->nmeaStrings[_gps_state->nmeaStringIndex].length = length;
	_gps_state->nmeaStrings[_gps_state->nmeaStringIndex].timestamp = timestamp;
		
	// advance to the next sentence
	_gps_state->nmeaStringIndex = (_gps_state->nmeaStringIndex + 1) % 7;
	pthread_mutex_unlock(&nmea_lock);	
}

/**
 * Callback to inform framework of the GPS engine's capabilities.
 * The GPS driver is receive-only, and does not support assisted GPS (aGPS)
 */ 
static void set_capabilities_callback(uint32_t capabilities)
{
    checkAndClearExceptionFromCallback(__FUNCTION__);
}

/**
 * Callback utility for acquiring the GPS wakelock.
 * This can be used to prevent the CPU from suspending while handling GPS events.
 */ 
static void acquire_wakelock_callback()
{
	checkAndClearExceptionFromCallback(__FUNCTION__);
}

/**
 * Callback utility for releasing the GPS wakelock.
 */ 
static void release_wakelock_callback()
{
	checkAndClearExceptionFromCallback(__FUNCTION__);
}

/**
 * Callback for creating threads.
 * This must be used to create any threads that report events up to the framework.
 */ 
static pthread_t create_thread_callback(const char* name, void (*start)(void *), void* arg)
{	
	checkAndClearExceptionFromCallback(__FUNCTION__);
	return  pthread_create((pthread_t *)&arg, NULL, start, arg);
}

/*
 * Callback for requesting NTP time
 * Currently the GPS driver does not have a mechanism for returning it.
 */ 
static void request_utc_time_callback()
{
  checkAndClearExceptionFromCallback(__FUNCTION__);
}

/**
 * GPS callback structure
 */
GpsCallbacks sGpsCallbacks = {
    sizeof(GpsCallbacks),
    location_callback,
    status_callback,
    sv_status_callback,
    nmea_callback,
    set_capabilities_callback,
    acquire_wakelock_callback,
    release_wakelock_callback,
    create_thread_callback,
    request_utc_time_callback,
};

/**
 * Returns a GPS interface
 */  
static const GpsInterface* GetGpsInterface() 
{
	checkAndClearExceptionFromCallback(__FUNCTION__);

	if (!sGpsInterface) {
			sGpsInterface = gps_get_interface();
			if (!sGpsInterface || sGpsInterface->init(&sGpsCallbacks) != 0) {
					sGpsInterface = NULL;
					return NULL;
			}
	}
	
	return sGpsInterface;
}

/**
 * Closes the GPS interface.
 */ 
static void GpsLocationProvider_cleanup()
{
    const GpsInterface* interface = GetGpsInterface();
    if (interface)
        interface->cleanup();
	checkAndClearExceptionFromCallback(__FUNCTION__);
}

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-function"

/**
 * Sets the GPS position mode.
 * Unsupported for the GPS 1.0 driver.
*/
static int GpsLocationProvider_set_position_mode(int mode, 
	int recurrence, int min_interval, int preferred_accuracy, 
	int preferred_time)
{
	checkAndClearExceptionFromCallback(__FUNCTION__);
	
	const GpsInterface* interface = GetGpsInterface();

	if (interface)
			return (interface->set_position_mode(mode, recurrence, min_interval, preferred_accuracy, preferred_time) == 0);
	else
			return 0;
}

/**
 * Stops GPS interface.
*/ 
static int GpsLocationProvider_stop()
{
	checkAndClearExceptionFromCallback(__FUNCTION__);
	
	const GpsInterface* interface = GetGpsInterface();	
	
	if (interface)
			return (interface->stop() == 0);
	else
			return 0;
}
#pragma GCC diagnostic pop

/**
 * Starts GPS interface
 */ 
static int GpsLocationProvider_start()
{
	checkAndClearExceptionFromCallback(__FUNCTION__);
	
  const GpsInterface* interface = GetGpsInterface();

	if (interface)
			return (interface->start() == 0);
	else
			return 0;
}

/**
 * Obtains a GpsInterface reference
 */ 
static int GpsLocationProvider_init()
{
	checkAndClearExceptionFromCallback(__FUNCTION__);
	
	const GpsInterface* interface = GetGpsInterface();
	
	if (!interface)
			return 0;
	
	if (!sGpsDebugInterface)
		 sGpsDebugInterface = (const GpsDebugInterface*)interface->get_extension(GPS_DEBUG_INTERFACE);

	return 1;
}

/**
 * Helper function for providing 
 * descriptive GPS status
 */  
char* getGpsStatusDefines(uint16_t value)
{
	switch (value)
	{
		case GPS_STATUS_NONE:
			return "GPS status unknown";
			
		case GPS_STATUS_SESSION_BEGIN:
			return "GPS is navigating";
			
		case GPS_STATUS_SESSION_END:
			return "GPS has stopped navigating";						
			
		case GPS_STATUS_ENGINE_ON:
			return "GPS has powered on but is not navigating";						
			
		case GPS_STATUS_ENGINE_OFF:
			return "GPS is powered off";												
	}
	
	return "Unknown";
}

/**
 * Display Thread.
 */
static void*
gps_display_thread( void*  arg )
{
	checkAndClearExceptionFromCallback(__FUNCTION__);
	
	int i = 0;
	int lockStatus;
	GpsUtcTime GPS_timestamp;
	struct timespec local_timeStamp;
	int ret;
	const uint TIME_FMT = strlen("2012-12-31 12:59:59.123456789") + 1;
    	char timestr[TIME_FMT];


	while (_gps_state->runThread) {
		// clear the output screen
		system("tput clear");
		
		printf("======================================================================================================\n");
		printf("GPS         Status\n");
		printf("            ==========================================================================================\n");	
				
		pthread_mutex_lock(&status_lock);
		printf("            %s\n", getGpsStatusDefines(lockStatus));
		pthread_mutex_unlock(&status_lock);			
		
				
		printf("======================================================================================================\n");
		printf("Location    Flags     Latitude    Longitude       Altitude      Speed        Accuracy    Timestamp\n");
		printf("            ==========================================================================================\n");	
		
		pthread_mutex_lock(&location_lock);
		printf("          %4d       %10f   %10f	  %10f  %10f   %10f    %lld\n", 
				_gps_state->location.flags, 
				_gps_state->location.latitude, 
				_gps_state->location.longitude, 
				_gps_state->location.altitude, 
				_gps_state->location.speed, 
				_gps_state->location.accuracy, 
				_gps_state->location.timestamp);
		GPS_timestamp = _gps_state->location.timestamp;
		pthread_mutex_unlock(&location_lock);				
		
		printf("======================================================================================================\n");
		printf("Satellite   ID        SnR         Elevation       Azimuth\n");	
		printf("            ==========================================================================================\n");	
		pthread_mutex_lock(&sv_lock);
		for (i = 0; i < _gps_state->sv_status.num_svs; ++i)
		{
			printf("          %4d       %10f  %10f       %10f \n", 
			_gps_state->sv_status.sv_list[i].prn, 
			_gps_state->sv_status.sv_list[i].snr, 
			_gps_state->sv_status.sv_list[i].elevation, 
			_gps_state->sv_status.sv_list[i].azimuth);
		}	
		pthread_mutex_unlock(&sv_lock);
		
		printf("======================================================================================================\n");
		printf("NMEA        Timestamp             String\n");
		printf("            ==========================================================================================\n");	
		
		pthread_mutex_lock(&nmea_lock);
		for (i = 0; i < sizeof(_gps_state->nmeaStrings)/sizeof(NmeaString); ++i)
		{
			printf("            %lld                     %s", _gps_state->nmeaStrings[i].timestamp, _gps_state->nmeaStrings[i].string);
		}
		pthread_mutex_unlock(&nmea_lock);

		// Check if we need to act on the timestamp
		if (1 == setTimeAndExit)
		{
			if (0 != GPS_timestamp)
                        {
                                local_timeStamp.tv_sec = (GPS_timestamp / 1000);
                                local_timeStamp.tv_nsec = 0;

                                ret = clock_settime(CLOCK_REALTIME, &local_timeStamp);
                                if (0 == ret)
				{
					if (0 == timespec2str(timestr, TIME_FMT, &local_timeStamp))
       						printf("time: %s\n", timestr);

                                        printf("Clock updated.\n");
				}
                                else
                                        printf("Clock update failed: %d\n", ret);

                                // We're done:
                                _gps_state->runThread = 0;
                        }
			else
			{
				printf("Waiting for GPS timestamp...\n");
			}
		}



		// display every 1 second
		sleep(1);		
	}
	
Exit:
  return NULL;
}

/*
 * Starts up the display thread.
 */
static void
StartDisplayThread( GpsState*  state )
{
	checkAndClearExceptionFromCallback(__FUNCTION__);
	
	state->nmeaStringIndex 	= 0;

	if ( pthread_create( &state->thread, NULL, gps_display_thread, state ) != 0 ) {
			printf("could not create gps display thread: %s\n", strerror(errno));
			goto Fail;
	}

Fail:
	return;
}

/*
 * GPS test application.
 * This application demonstrates how to use the GPS Driver in Logic PD's Torpedo WiFi.
 * The application sets up the callbacks required by the GPS driver to call into.
 * The callbacks simply store the GPS data.
 * A display thread displays the stored the data, based on one Hertz.
 * To exit the application, simply press CTRL+C.		
 */ 
int main(int argc, char **argv)
{
	int userInput = '0';
	int c, index = 0;

	while ((c = getopt (argc, argv, "t")) != -1)
	{
		switch(c)
		{
			case 't':
				setTimeAndExit = 1;
				break;
		}
	}

			  
	// Start up the display thread
	GpsState*  s = _gps_state;
	s->runThread = 1;
	StartDisplayThread(s);

	// Initialize the GPS driver reference
	GpsLocationProvider_init();

	// Start up the GPS interface
	GpsLocationProvider_start();
	
	userInput = getchar();
	
	if (userInput == EOF)
	{
		s->runThread = 0;
		
		// Stop GPS interface
		GpsLocationProvider_cleanup();
		
		return 0;
	}
}
